unit ISP;

interface

uses Ole2, OleCtl, Classes, Graphics, OleCtrls;

const

{ NotificationModeConstants }

  icCompleteMode = 0;
  icContinousMode = 1;

{ StateConstants }

  prcConnecting = 1;
  prcResolvingHost = 2;
  prcHostResolved = 3;
  prcConnected = 4;
  prcDisconnecting = 5;
  prcDisconnected = 6;

{ TimeoutConstants }

  prcConnectTimeout = 1;
  prcReceiveTimeout = 2;
  prcUserTimeout = 65;

{ MethodConstants }

  prcGet = 1;
  prcHead = 2;
  prcPost = 3;
  prcPut = 4;

{ FTPTypeConstants }

  ftpAscii = 0;
  ftpEBCDIC = 1;
  ftpImage = 2;
  ftpBinary = 3;

{ FTPModeConstants }

  ftpStream = 0;
  ftpBlock = 1;
  ftpCompressed = 2;

{ FTPOperationConstants }

  ftpFile = 0;
  ftpList = 1;
  ftpNameList = 2;

{ FTPProtocolStateConstants }

  ftpBase = 0;
  ftpAuthentication = 1;
  ftpTransaction = 2;

{ HTTPProtocolStateConstants }

  httpBase = 0;
  httpTransferring = 1;

{ NNTPProtocolStateConstants }

  nntpBase = 0;
  nntpTransaction = 1;
  
{ PopProtocolStateConstants }

  popBase = 0;
  popAuthorization = 1;
  popTransaction = 2;
  popUpdate = 3;

{ SMTPProtocolStateConstants }

  smtpBase = 0;
  smtpTransaction = 1;

{ SCKSTATEConstants }

  sckClosed = 0;
  sckOpen = 1;
  sckListening = 2;
  sckConnectionPending = 3;
  sckResolvingHost = 4;
  sckHostResolved = 5;
  sckConnecting = 6;
  sckConnected = 7;
  sckClosing = 8;
  sckError = 9;

{ DocInput/Output Constants } 

  icDocNone = 0;
  icDocBegin = 1;
  icDocHeaders = 2;
  icDocData = 3;
  icDocError = 4;
  icDocEnd = 5;
  
type

{ TFTP }

  TFTPError = procedure(Sender: TObject; Number: Smallint; var Description: string; Scode: Integer; const Source, HelpFile: string; HelpContext: Integer; var CancelDisplay: TOleBool) of object;
  TFTPTimeout = procedure(Sender: TObject; event: Smallint; var Continue: TOleBool) of object;
  TFTPStateChanged = procedure(Sender: TObject; State: Smallint) of object;
  TFTPProtocolStateChanged = procedure(Sender: TObject; ProtocolState: Smallint) of object;
  TFTPBusy = procedure(Sender: TObject; isBusy: TOleBool) of object;
  TFTPDocInput = procedure(Sender: TObject; const DocInput: Variant) of object;
  TFTPDocOutput = procedure(Sender: TObject; const DocOutput: Variant) of object;
  TFTPListItem = procedure(Sender: TObject; const Item: Variant) of object;

  TFTP = class(TOleControl)
  private
    FOnError: TFTPError;
    FOnTimeout: TFTPTimeout;
    FOnCancel: TNotifyEvent;
    FOnStateChanged: TFTPStateChanged;
    FOnProtocolStateChanged: TFTPProtocolStateChanged;
    FOnBusy: TFTPBusy;
    FOnLog: TNotifyEvent;
    FOnDocInput: TFTPDocInput;
    FOnDocOutput: TFTPDocOutput;
    FOnAbort: TNotifyEvent;
    FOnAccount: TNotifyEvent;
    FOnChangeDir: TNotifyEvent;
    FOnCreateDir: TNotifyEvent;
    FOnDeleteDir: TNotifyEvent;
    FOnDelFile: TNotifyEvent;
    FOnHelp: TNotifyEvent;
    FOnmode: TNotifyEvent;
    FOnNoop: TNotifyEvent;
    FOnParentDir: TNotifyEvent;
    FOnPrintDir: TNotifyEvent;
    FOnExecute: TNotifyEvent;
    FOnStatus: TNotifyEvent;
    FOnReinitialize: TNotifyEvent;
    FOnSystem: TNotifyEvent;
    FOnSite: TNotifyEvent;
    FOnType: TNotifyEvent;
    FOnListItem: TFTPListItem;
    function Get_Timeout(event: Smallint): Integer; stdcall;
    procedure Set_Timeout(event: Smallint; Value: Integer); stdcall;
    procedure Set_EnableTimer(event: Smallint; Value: TOleBool); stdcall;
  protected
    procedure InitControlData; override;
  public
    procedure AboutBox; stdcall;
    procedure Cancel; stdcall;
    procedure Connect(const RemoteHost, RemotePort: Variant); stdcall;
    procedure Authenticate(const UserId, Password: Variant); stdcall;
    procedure SendDoc(const URL, Headers, InputData, InputFile, OutputFile: Variant); stdcall;
    procedure GetDoc(const URL, Headers, OutputFile: Variant); stdcall;
    procedure Abort; stdcall;
    procedure Account(const Account: string); stdcall;
    procedure ChangeDir(const directory: string); stdcall;
    procedure CreateDir(const directory: string); stdcall;
    procedure DeleteDir(const directory: string); stdcall;
    procedure DeleteFile(const FileName: string); stdcall;
    procedure Quit; stdcall;
    procedure Help(const Help: string); stdcall;
    procedure Noop; stdcall;
    procedure mode(ftpMode: TOleEnum); stdcall;
    procedure Type_(ftpType: TOleEnum); stdcall;
    procedure List(const List: string); stdcall;
    procedure NameList(const NameList: string); stdcall;
    procedure ParentDir; stdcall;
    procedure PrintDir; stdcall;
    procedure Execute(const Execute: string); stdcall;
    procedure Status(const Status: string); stdcall;
    procedure PutFile(const srcFileName, destFileName: string); stdcall;
    procedure Reinitialize; stdcall;
    procedure System; stdcall;
    procedure GetFile(const srcFileName, destFileName: string); stdcall;
    procedure PutFileUnique(const FileName: string); stdcall;
    procedure Site(const Site: string); stdcall;
    property State: Smallint index 503 read GetSmallintProp;
    property ProtocolState: Smallint index 504 read GetSmallintProp;
    property ReplyString: string index 505 read GetStringProp;
    property ReplyCode: Integer index 506 read GetIntegerProp;
    property Errors: Variant index 508 read GetVariantProp;
    property Busy: TOleBool index 509 read GetOleBoolProp;
    property StateString: string index 511 read GetStringProp;
    property ProtocolStateString: string index 512 read GetStringProp;
    property DocInput: Variant index 1002 read GetVariantProp;
    property DocOutput: Variant index 1003 read GetVariantProp;
    property Operation: TOleEnum index 5 read GetOleEnumProp;
    property Timeout[event: Smallint]: Integer read Get_Timeout write Set_Timeout;
    property EnableTimer[event: Smallint]: TOleBool write Set_EnableTimer;
  published
    property RemoteHost: string index 0 read GetStringProp write SetStringProp stored False;
    property RemotePort: Integer index 502 read GetIntegerProp write SetIntegerProp stored False;
    property NotificationMode: Smallint index 510 read GetSmallintProp write SetSmallintProp stored False;
    property Logging: TOleBool index 514 read GetOleBoolProp write SetOleBoolProp stored False;
    property UserId: string index 601 read GetStringProp write SetStringProp stored False;
    property Password: string index 602 read GetStringProp write SetStringProp stored False;
    property URL: string index 1001 read GetStringProp write SetStringProp stored False;
    property AppendToFile: TOleBool index 1 read GetOleBoolProp write SetOleBoolProp stored False;
    property ListItemNotify: TOleBool index 2 read GetOleBoolProp write SetOleBoolProp stored False;
    property RemoteFile: string index 3 read GetStringProp write SetStringProp stored False;
    property OnError: TFTPError read FOnError write FOnError;
    property OnTimeout: TFTPTimeout read FOnTimeout write FOnTimeout;
    property OnCancel: TNotifyEvent read FOnCancel write FOnCancel;
    property OnStateChanged: TFTPStateChanged read FOnStateChanged write FOnStateChanged;
    property OnProtocolStateChanged: TFTPProtocolStateChanged read FOnProtocolStateChanged write FOnProtocolStateChanged;
    property OnBusy: TFTPBusy read FOnBusy write FOnBusy;
    property OnLog: TNotifyEvent read FOnLog write FOnLog;
    property OnDocInput: TFTPDocInput read FOnDocInput write FOnDocInput;
    property OnDocOutput: TFTPDocOutput read FOnDocOutput write FOnDocOutput;
    property OnAbort: TNotifyEvent read FOnAbort write FOnAbort;
    property OnAccount: TNotifyEvent read FOnAccount write FOnAccount;
    property OnChangeDir: TNotifyEvent read FOnChangeDir write FOnChangeDir;
    property OnCreateDir: TNotifyEvent read FOnCreateDir write FOnCreateDir;
    property OnDeleteDir: TNotifyEvent read FOnDeleteDir write FOnDeleteDir;
    property OnDelFile: TNotifyEvent read FOnDelFile write FOnDelFile;
    property OnHelp: TNotifyEvent read FOnHelp write FOnHelp;
    property Onmode: TNotifyEvent read FOnmode write FOnmode;
    property OnNoop: TNotifyEvent read FOnNoop write FOnNoop;
    property OnParentDir: TNotifyEvent read FOnParentDir write FOnParentDir;
    property OnPrintDir: TNotifyEvent read FOnPrintDir write FOnPrintDir;
    property OnExecute: TNotifyEvent read FOnExecute write FOnExecute;
    property OnStatus: TNotifyEvent read FOnStatus write FOnStatus;
    property OnReinitialize: TNotifyEvent read FOnReinitialize write FOnReinitialize;
    property OnSystem: TNotifyEvent read FOnSystem write FOnSystem;
    property OnSite: TNotifyEvent read FOnSite write FOnSite;
    property OnType: TNotifyEvent read FOnType write FOnType;
    property OnListItem: TFTPListItem read FOnListItem write FOnListItem;
  end;

{ THTML }

  THTMLError = procedure(Sender: TObject; Number: Smallint; var Description: string; Scode: Integer; const Source, HelpFile: string; HelpContext: Integer; var CancelDisplay: TOleBool) of object;
  THTMLDocInput = procedure(Sender: TObject; const DocInput: Variant) of object;
  THTMLDocOutput = procedure(Sender: TObject; const DocOutput: Variant) of object;
  THTMLDoRequestDoc = procedure(Sender: TObject; const URL: string; const Element, DocInput: Variant; var EnableDefault: TOleBool) of object;
  THTMLDoRequestEmbedded = procedure(Sender: TObject; const URL: string; const Element, DocInput: Variant; var EnableDefault: TOleBool) of object;
  THTMLDoRequestSubmit = procedure(Sender: TObject; const URL: string; const Form, DocOutput: Variant; var EnableDefault: TOleBool) of object;
  THTMLDoNewElement = procedure(Sender: TObject; const ElemType: string; EndTag: TOleBool; const Attrs: Variant; const Text: string; var EnableDefault: TOleBool) of object;

  THTML = class(TOleControl)
  private
    FOnError: THTMLError;
    FOnDocInput: THTMLDocInput;
    FOnDocOutput: THTMLDocOutput;
    FOnParseComplete: TNotifyEvent;
    FOnLayoutComplete: TNotifyEvent;
    FOnTimeout: TNotifyEvent;
    FOnBeginRetrieval: TNotifyEvent;
    FOnUpdateRetrieval: TNotifyEvent;
    FOnEndRetrieval: TNotifyEvent;
    FOnDoRequestDoc: THTMLDoRequestDoc;
    FOnDoRequestEmbedded: THTMLDoRequestEmbedded;
    FOnDoRequestSubmit: THTMLDoRequestSubmit;
    FOnDoNewElement: THTMLDoNewElement;
  protected
    procedure InitControlData; override;
  public
    procedure AboutBox; stdcall;
    procedure RequestDoc(const URL: string); stdcall;
    procedure RequestAllEmbedded; stdcall;
    procedure Cancel(const Message: Variant); stdcall;
    property DocInput: Variant index 1002 read GetVariantProp;
    property DocOutput: Variant index 1003 read GetVariantProp;
    property URL: string index 1001 read GetStringProp;
    property RequestURL: string index 2 read GetStringProp;
    property BaseURL: string index 3 read GetStringProp;
    property Forms: Variant index 4 read GetVariantProp;
    property TotalWidth: Integer index 5 read GetIntegerProp;
    property TotalHeight: Integer index 6 read GetIntegerProp;
    property RetrieveBytesTotal: Integer index 7 read GetIntegerProp;
    property RetrieveBytesDone: Integer index 8 read GetIntegerProp;
    property ParseDone: TOleBool index 9 read GetOleBoolProp;
    property LayoutDone: TOleBool index 10 read GetOleBoolProp;
    property SourceText: string index 14 read GetStringProp;
    property DocBackColor: TColor index 23 read GetColorProp;
    property DocForeColor: TColor index 24 read GetColorProp;
    property DocLinkColor: TColor index 25 read GetColorProp;
    property DocVisitedColor: TColor index 26 read GetColorProp;
    property Errors: Variant index 508 read GetVariantProp;
  published
    property Align;
    property ParentColor;
    property ParentFont;
    property TabStop;
    property DragCursor;
    property DragMode;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property Visible;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnStartDrag;
    property OnClick;
    property OnDblClick;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property DeferRetrieval: TOleBool index 11 read GetOleBoolProp write SetOleBoolProp stored False;
    property ViewSource: TOleBool index 12 read GetOleBoolProp write SetOleBoolProp stored False;
    property RetainSource: TOleBool index 13 read GetOleBoolProp write SetOleBoolProp stored False;
    property ElemNotification: TOleBool index 15 read GetOleBoolProp write SetOleBoolProp stored False;
    property Timeout: Integer index 507 read GetIntegerProp write SetIntegerProp stored False;
    property Redraw: TOleBool index 17 read GetOleBoolProp write SetOleBoolProp stored False;
    property UnderlineLinks: TOleBool index 18 read GetOleBoolProp write SetOleBoolProp stored False;
    property UseDocColors: TOleBool index 19 read GetOleBoolProp write SetOleBoolProp stored False;
    property BackImage: string index 20 read GetStringProp write SetStringProp stored False;
    property BackColor: TColor index -501 read GetColorProp write SetColorProp stored False;
    property ForeColor: TColor index -513 read GetColorProp write SetColorProp stored False;
    property LinkColor: TColor index 21 read GetColorProp write SetColorProp stored False;
    property VisitedColor: TColor index 22 read GetColorProp write SetColorProp stored False;
    property Font: Variant index -512 read GetVariantProp write SetVariantProp stored False;
    property FixedFont: Variant index 27 read GetVariantProp write SetVariantProp stored False;
    property Heading1Font: Variant index 28 read GetVariantProp write SetVariantProp stored False;
    property Heading2Font: Variant index 29 read GetVariantProp write SetVariantProp stored False;
    property Heading3Font: Variant index 30 read GetVariantProp write SetVariantProp stored False;
    property Heading4Font: Variant index 31 read GetVariantProp write SetVariantProp stored False;
    property Heading5Font: Variant index 32 read GetVariantProp write SetVariantProp stored False;
    property Heading6Font: Variant index 33 read GetVariantProp write SetVariantProp stored False;
    property OnError: THTMLError read FOnError write FOnError;
    property OnDocInput: THTMLDocInput read FOnDocInput write FOnDocInput;
    property OnDocOutput: THTMLDocOutput read FOnDocOutput write FOnDocOutput;
    property OnParseComplete: TNotifyEvent read FOnParseComplete write FOnParseComplete;
    property OnLayoutComplete: TNotifyEvent read FOnLayoutComplete write FOnLayoutComplete;
    property OnTimeout: TNotifyEvent read FOnTimeout write FOnTimeout;
    property OnBeginRetrieval: TNotifyEvent read FOnBeginRetrieval write FOnBeginRetrieval;
    property OnUpdateRetrieval: TNotifyEvent read FOnUpdateRetrieval write FOnUpdateRetrieval;
    property OnEndRetrieval: TNotifyEvent read FOnEndRetrieval write FOnEndRetrieval;
    property OnDoRequestDoc: THTMLDoRequestDoc read FOnDoRequestDoc write FOnDoRequestDoc;
    property OnDoRequestEmbedded: THTMLDoRequestEmbedded read FOnDoRequestEmbedded write FOnDoRequestEmbedded;
    property OnDoRequestSubmit: THTMLDoRequestSubmit read FOnDoRequestSubmit write FOnDoRequestSubmit;
    property OnDoNewElement: THTMLDoNewElement read FOnDoNewElement write FOnDoNewElement;
  end; 
  
{ THTTP }

  THTTPError = procedure(Sender: TObject; Number: Smallint; var Description: string; Scode: Integer; const Source, HelpFile: string; HelpContext: Integer; var CancelDisplay: TOleBool) of object;
  THTTPTimeout = procedure(Sender: TObject; event: Smallint; var Continue: TOleBool) of object;
  THTTPStateChanged = procedure(Sender: TObject; State: Smallint) of object;
  THTTPProtocolStateChanged = procedure(Sender: TObject; ProtocolState: Smallint) of object;
  THTTPBusy = procedure(Sender: TObject; isBusy: TOleBool) of object;
  THTTPDocInput = procedure(Sender: TObject; const DocInput: Variant) of object;
  THTTPDocOutput = procedure(Sender: TObject; const DocOutput: Variant) of object;

  THTTP = class(TOleControl)
  private
    FOnError: THTTPError;
    FOnTimeout: THTTPTimeout;
    FOnCancel: TNotifyEvent;
    FOnStateChanged: THTTPStateChanged;
    FOnProtocolStateChanged: THTTPProtocolStateChanged;
    FOnBusy: THTTPBusy;
    FOnLog: TNotifyEvent;
    FOnDocInput: THTTPDocInput;
    FOnDocOutput: THTTPDocOutput;
    function Get_Timeout(event: Smallint): Integer; stdcall;
    procedure Set_Timeout(event: Smallint; Value: Integer); stdcall;
    procedure Set_EnableTimer(event: Smallint; Value: TOleBool); stdcall;
  protected
    procedure InitControlData; override;
  public
    procedure AboutBox; stdcall;
    procedure Cancel; stdcall;
    procedure SendDoc(const URL, Headers, InputData, InputFile, OutputFile: Variant); stdcall;
    procedure GetDoc(const URL, Headers, OutputFile: Variant); stdcall;
    property State: Smallint index 503 read GetSmallintProp;
    property ProtocolState: Smallint index 504 read GetSmallintProp;
    property ReplyString: string index 505 read GetStringProp;
    property ReplyCode: Integer index 506 read GetIntegerProp;
    property Errors: Variant index 508 read GetVariantProp;
    property Busy: TOleBool index 509 read GetOleBoolProp;
    property StateString: string index 511 read GetStringProp;
    property ProtocolStateString: string index 512 read GetStringProp;
    property DocInput: Variant index 1002 read GetVariantProp;
    property DocOutput: Variant index 1003 read GetVariantProp;
    property Timeout[event: Smallint]: Integer read Get_Timeout write Set_Timeout;
    property EnableTimer[event: Smallint]: TOleBool write Set_EnableTimer;
  published
    property RemoteHost: string index 0 read GetStringProp write SetStringProp stored False;
    property RemotePort: Integer index 502 read GetIntegerProp write SetIntegerProp stored False;
    property NotificationMode: Smallint index 510 read GetSmallintProp write SetSmallintProp stored False;
    property Logging: TOleBool index 514 read GetOleBoolProp write SetOleBoolProp stored False;
    property Document: string index 10 read GetStringProp write SetStringProp stored False;
    property Method: TOleEnum index 11 read GetOleEnumProp write SetOleEnumProp stored False;
    property URL: string index 1001 read GetStringProp write SetStringProp stored False;
    property OnError: THTTPError read FOnError write FOnError;
    property OnTimeout: THTTPTimeout read FOnTimeout write FOnTimeout;
    property OnCancel: TNotifyEvent read FOnCancel write FOnCancel;
    property OnStateChanged: THTTPStateChanged read FOnStateChanged write FOnStateChanged;
    property OnProtocolStateChanged: THTTPProtocolStateChanged read FOnProtocolStateChanged write FOnProtocolStateChanged;
    property OnBusy: THTTPBusy read FOnBusy write FOnBusy;
    property OnLog: TNotifyEvent read FOnLog write FOnLog;
    property OnDocInput: THTTPDocInput read FOnDocInput write FOnDocInput;
    property OnDocOutput: THTTPDocOutput read FOnDocOutput write FOnDocOutput;
  end;

{ TNNTP }
  
  TNNTPError = procedure(Sender: TObject; Number: Smallint; var Description: string; Scode: Integer; const Source, HelpFile: string; HelpContext: Integer; var CancelDisplay: TOleBool) of object;
  TNNTPTimeout = procedure(Sender: TObject; event: Smallint; var Continue: TOleBool) of object;
  TNNTPStateChanged = procedure(Sender: TObject; State: Smallint) of object;
  TNNTPProtocolStateChanged = procedure(Sender: TObject; ProtocolState: Smallint) of object;
  TNNTPBusy = procedure(Sender: TObject; isBusy: TOleBool) of object;
  TNNTPDocInput = procedure(Sender: TObject; const DocInput: Variant) of object;
  TNNTPDocOutput = procedure(Sender: TObject; const DocOutput: Variant) of object;
  TNNTPAuthenticateRequest = procedure(Sender: TObject; var UserId, Password: string) of object;
  TNNTPAuthenticateResponse = procedure(Sender: TObject; Authenticated: TOleBool) of object;
  TNNTPBanner = procedure(Sender: TObject; const Banner: string) of object;
  TNNTPSelectGroup = procedure(Sender: TObject; const groupName: string; firstMessage, lastMessage, msgCount: Integer) of object;
  TNNTPNextArticle = procedure(Sender: TObject; articleNumber: Integer; const messageID: string) of object;
  TNNTPlastArticle = procedure(Sender: TObject; articleNumber: Integer; const messageID: string) of object;
  TNNTPArticleStatus = procedure(Sender: TObject; articleNumber: Integer; const messageID: string) of object;

  TNNTP = class(TOleControl)
  private
    FOnError: TNNTPError;
    FOnTimeout: TNNTPTimeout;
    FOnCancel: TNotifyEvent;
    FOnStateChanged: TNNTPStateChanged;
    FOnProtocolStateChanged: TNNTPProtocolStateChanged;
    FOnBusy: TNNTPBusy;
    FOnLog: TNotifyEvent;
    FOnDocInput: TNNTPDocInput;
    FOnDocOutput: TNNTPDocOutput;
    FOnAuthenticateRequest: TNNTPAuthenticateRequest;
    FOnAuthenticateResponse: TNNTPAuthenticateResponse;
    FOnBanner: TNNTPBanner;
    FOnSelectGroup: TNNTPSelectGroup;
    FOnNextArticle: TNNTPNextArticle;
    FOnlastArticle: TNNTPlastArticle;
    FOnArticleStatus: TNNTPArticleStatus;
    function Get_Timeout(event: Smallint): Integer; stdcall;
    procedure Set_Timeout(event: Smallint; Value: Integer); stdcall;
    procedure Set_EnableTimer(event: Smallint; Value: TOleBool); stdcall;
  protected
    procedure InitControlData; override;
  public
    procedure AboutBox; stdcall;
    procedure Cancel; stdcall;
    procedure Connect(const RemoteHost, RemotePort: Variant); stdcall;
    procedure SendDoc(const URL, Headers, InputData, InputFile, OutputFile: Variant); stdcall;
    procedure GetDoc(const URL, Headers, OutputFile: Variant); stdcall;
    procedure GetAdministrationFile(const LastUpdate: Variant); stdcall;
    procedure SelectGroup(const groupName: string); stdcall;
    procedure SetNextArticle; stdcall;
    procedure SetLastArticle; stdcall;
    procedure GetArticleNumbers(const groupName: Variant); stdcall;
    procedure GetArticleHeaders(const header: string; const firstArticle, lastArticle: Variant); stdcall;
    procedure GetArticleByArticleNumber(const articleNumber: Variant); stdcall;
    procedure GetArticleByMessageID(const messageID: string); stdcall;
    procedure GetHeaderByArticleNumber(const articleNumber: Variant); stdcall;
    procedure GetHeaderByMessageID(const messageID: string); stdcall;
    procedure GetBodyByArticleNumber(const articleNumber: Variant); stdcall;
    procedure GetBodyByMessageID(const messageID: string); stdcall;
    procedure GetStatByArticleNumber(const articleNumber: Variant); stdcall;
    procedure GetOverviewFormat; stdcall;
    procedure GetOverview(const firstArticle, lastArticle: Variant); stdcall;
    procedure ListGroups; stdcall;
    procedure ListGroupDescriptions; stdcall;
    procedure ListNewGroups(const LastUpdate: Variant); stdcall;
    procedure Quit; stdcall;
    property State: Smallint index 503 read GetSmallintProp;
    property ProtocolState: Smallint index 504 read GetSmallintProp;
    property ReplyString: string index 505 read GetStringProp;
    property ReplyCode: Integer index 506 read GetIntegerProp;
    property Errors: Variant index 508 read GetVariantProp;
    property Busy: TOleBool index 509 read GetOleBoolProp;
    property StateString: string index 511 read GetStringProp;
    property ProtocolStateString: string index 512 read GetStringProp;
    property DocInput: Variant index 1002 read GetVariantProp;
    property DocOutput: Variant index 1003 read GetVariantProp;
    property ArticleNumbersSupported: TOleBool index 1 read GetOleBoolProp;
    property OverviewSupported: TOleBool index 2 read GetOleBoolProp;
    property PostingAllowed: TOleBool index 3 read GetOleBoolProp;
    property Timeout[event: Smallint]: Integer read Get_Timeout write Set_Timeout;
    property EnableTimer[event: Smallint]: TOleBool write Set_EnableTimer;
  published
    property RemoteHost: string index 0 read GetStringProp write SetStringProp stored False;
    property RemotePort: Integer index 502 read GetIntegerProp write SetIntegerProp stored False;
    property NotificationMode: Smallint index 510 read GetSmallintProp write SetSmallintProp stored False;
    property Logging: TOleBool index 514 read GetOleBoolProp write SetOleBoolProp stored False;
    property URL: string index 1001 read GetStringProp write SetStringProp stored False;
    property LastUpdate: TOleDate index 4 read GetOleDateProp write SetOleDateProp stored False;
    property OnError: TNNTPError read FOnError write FOnError;
    property OnTimeout: TNNTPTimeout read FOnTimeout write FOnTimeout;
    property OnCancel: TNotifyEvent read FOnCancel write FOnCancel;
    property OnStateChanged: TNNTPStateChanged read FOnStateChanged write FOnStateChanged;
    property OnProtocolStateChanged: TNNTPProtocolStateChanged read FOnProtocolStateChanged write FOnProtocolStateChanged;
    property OnBusy: TNNTPBusy read FOnBusy write FOnBusy;
    property OnLog: TNotifyEvent read FOnLog write FOnLog;
    property OnDocInput: TNNTPDocInput read FOnDocInput write FOnDocInput;
    property OnDocOutput: TNNTPDocOutput read FOnDocOutput write FOnDocOutput;
    property OnAuthenticateRequest: TNNTPAuthenticateRequest read FOnAuthenticateRequest write FOnAuthenticateRequest;
    property OnAuthenticateResponse: TNNTPAuthenticateResponse read FOnAuthenticateResponse write FOnAuthenticateResponse;
    property OnBanner: TNNTPBanner read FOnBanner write FOnBanner;
    property OnSelectGroup: TNNTPSelectGroup read FOnSelectGroup write FOnSelectGroup;
    property OnNextArticle: TNNTPNextArticle read FOnNextArticle write FOnNextArticle;
    property OnlastArticle: TNNTPlastArticle read FOnlastArticle write FOnlastArticle;
    property OnArticleStatus: TNNTPArticleStatus read FOnArticleStatus write FOnArticleStatus;
  end;

{ TPOP }
  
  TPOPError = procedure(Sender: TObject; Number: Smallint; var Description: string; Scode: Integer; const Source, HelpFile: string; HelpContext: Integer; var CancelDisplay: TOleBool) of object;
  TPOPTimeout = procedure(Sender: TObject; event: Smallint; var Continue: TOleBool) of object;
  TPOPStateChanged = procedure(Sender: TObject; State: Smallint) of object;
  TPOPProtocolStateChanged = procedure(Sender: TObject; ProtocolState: Smallint) of object;
  TPOPBusy = procedure(Sender: TObject; isBusy: TOleBool) of object;
  TPOPDocOutput = procedure(Sender: TObject; const DocOutput: Variant) of object;
  TPOPMessageSize = procedure(Sender: TObject; msgSize: Integer) of object;
  TPOPLast = procedure(Sender: TObject; Number: Integer) of object;
  TPOPRefreshMessageCount = procedure(Sender: TObject; Number: Integer) of object;

  TPOP = class(TOleControl)
  private
    FOnError: TPOPError;
    FOnTimeout: TPOPTimeout;
    FOnCancel: TNotifyEvent;
    FOnStateChanged: TPOPStateChanged;
    FOnProtocolStateChanged: TPOPProtocolStateChanged;
    FOnBusy: TPOPBusy;
    FOnLog: TNotifyEvent;
    FOnDocOutput: TPOPDocOutput;
    FOnMessageSize: TPOPMessageSize;
    FOnDelete: TNotifyEvent;
    FOnReset: TNotifyEvent;
    FOnLast: TPOPLast;
    FOnNoop: TNotifyEvent;
    FOnRefreshMessageCount: TPOPRefreshMessageCount;
    function Get_Timeout(event: Smallint): Integer; stdcall;
    procedure Set_Timeout(event: Smallint; Value: Integer); stdcall;
    procedure Set_EnableTimer(event: Smallint; Value: TOleBool); stdcall;
  protected
    procedure InitControlData; override;
  public
    procedure AboutBox; stdcall;
    procedure Cancel; stdcall;
    procedure Connect(const RemoteHost, RemotePort: Variant); stdcall;
    procedure Authenticate(const UserId, Password: Variant); stdcall;
    procedure GetDoc(const URL, Headers, OutputFile: Variant); stdcall;
    procedure MessageSize(MessageNumber: Smallint); stdcall;
    procedure RetrieveMessage(MessageNumber: Smallint); stdcall;
    procedure Delete(MessageNumber: Smallint); stdcall;
    procedure Reset; stdcall;
    procedure Last; stdcall;
    procedure Noop; stdcall;
    procedure TopMessage(MessageNumber: Smallint); stdcall;
    procedure Quit; stdcall;
    procedure RefreshMessageCount; stdcall;
    property State: Smallint index 503 read GetSmallintProp;
    property ProtocolState: Smallint index 504 read GetSmallintProp;
    property ReplyString: string index 505 read GetStringProp;
    property ReplyCode: Integer index 506 read GetIntegerProp;
    property Errors: Variant index 508 read GetVariantProp;
    property Busy: TOleBool index 509 read GetOleBoolProp;
    property StateString: string index 511 read GetStringProp;
    property ProtocolStateString: string index 512 read GetStringProp;
    property DocOutput: Variant index 1003 read GetVariantProp;
    property TopSupported: TOleBool index 2452 read GetOleBoolProp;
    property MessageCount: Smallint index 2454 read GetSmallintProp;
    property Timeout[event: Smallint]: Integer read Get_Timeout write Set_Timeout;
    property EnableTimer[event: Smallint]: TOleBool write Set_EnableTimer;
  published
    property RemoteHost: string index 0 read GetStringProp write SetStringProp stored False;
    property RemotePort: Integer index 502 read GetIntegerProp write SetIntegerProp stored False;
    property NotificationMode: Smallint index 510 read GetSmallintProp write SetSmallintProp stored False;
    property Logging: TOleBool index 514 read GetOleBoolProp write SetOleBoolProp stored False;
    property UserId: string index 601 read GetStringProp write SetStringProp stored False;
    property Password: string index 602 read GetStringProp write SetStringProp stored False;
    property URL: string index 1001 read GetStringProp write SetStringProp stored False;
    property TopLines: Integer index 2453 read GetIntegerProp write SetIntegerProp stored False;
    property OnError: TPOPError read FOnError write FOnError;
    property OnTimeout: TPOPTimeout read FOnTimeout write FOnTimeout;
    property OnCancel: TNotifyEvent read FOnCancel write FOnCancel;
    property OnStateChanged: TPOPStateChanged read FOnStateChanged write FOnStateChanged;
    property OnProtocolStateChanged: TPOPProtocolStateChanged read FOnProtocolStateChanged write FOnProtocolStateChanged;
    property OnBusy: TPOPBusy read FOnBusy write FOnBusy;
    property OnLog: TNotifyEvent read FOnLog write FOnLog;
    property OnDocOutput: TPOPDocOutput read FOnDocOutput write FOnDocOutput;
    property OnMessageSize: TPOPMessageSize read FOnMessageSize write FOnMessageSize;
    property OnDelete: TNotifyEvent read FOnDelete write FOnDelete;
    property OnReset: TNotifyEvent read FOnReset write FOnReset;
    property OnLast: TPOPLast read FOnLast write FOnLast;
    property OnNoop: TNotifyEvent read FOnNoop write FOnNoop;
    property OnRefreshMessageCount: TPOPRefreshMessageCount read FOnRefreshMessageCount write FOnRefreshMessageCount;
  end;

{ TSMTP }

  TSMTPError = procedure(Sender: TObject; Number: Smallint; var Description: string; Scode: Integer; const Source, HelpFile: string; HelpContext: Integer; var CancelDisplay: TOleBool) of object;
  TSMTPTimeout = procedure(Sender: TObject; event: Smallint; var Continue: TOleBool) of object;
  TSMTPStateChanged = procedure(Sender: TObject; State: Smallint) of object;
  TSMTPProtocolStateChanged = procedure(Sender: TObject; ProtocolState: Smallint) of object;
  TSMTPBusy = procedure(Sender: TObject; isBusy: TOleBool) of object;
  TSMTPDocInput = procedure(Sender: TObject; const DocInput: Variant) of object;

  TSMTP = class(TOleControl)
  private
    FOnError: TSMTPError;
    FOnTimeout: TSMTPTimeout;
    FOnCancel: TNotifyEvent;
    FOnStateChanged: TSMTPStateChanged;
    FOnProtocolStateChanged: TSMTPProtocolStateChanged;
    FOnBusy: TSMTPBusy;
    FOnLog: TNotifyEvent;
    FOnDocInput: TSMTPDocInput;
    FOnReset: TNotifyEvent;
    FOnVerify: TNotifyEvent;
    FOnExpand: TNotifyEvent;
    FOnHelp: TNotifyEvent;
    FOnNoop: TNotifyEvent;
    function Get_Timeout(event: Smallint): Integer; stdcall;
    procedure Set_Timeout(event: Smallint; Value: Integer); stdcall;
    procedure Set_EnableTimer(event: Smallint; Value: TOleBool); stdcall;
  protected
    procedure InitControlData; override;
  public
    procedure AboutBox; stdcall;
    procedure Cancel; stdcall;
    procedure Connect(const RemoteHost, RemotePort: Variant); stdcall;
    procedure SendDoc(const URL, Headers, InputData, InputFile, OutputFile: Variant); stdcall;
    procedure Reset; stdcall;
    procedure Verify(const name: string); stdcall;
    procedure Expand(const name: string); stdcall;
    procedure Help(const helpTopic: Variant); stdcall;
    procedure Noop; stdcall;
    procedure Quit; stdcall;
    property State: Smallint index 503 read GetSmallintProp;
    property ProtocolState: Smallint index 504 read GetSmallintProp;
    property ReplyString: string index 505 read GetStringProp;
    property ReplyCode: Integer index 506 read GetIntegerProp;
    property Errors: Variant index 508 read GetVariantProp;
    property Busy: TOleBool index 509 read GetOleBoolProp;
    property StateString: string index 511 read GetStringProp;
    property ProtocolStateString: string index 512 read GetStringProp;
    property DocInput: Variant index 1002 read GetVariantProp;
    property Timeout[event: Smallint]: Integer read Get_Timeout write Set_Timeout;
    property EnableTimer[event: Smallint]: TOleBool write Set_EnableTimer;
  published
    property RemoteHost: string index 0 read GetStringProp write SetStringProp stored False;
    property RemotePort: Integer index 502 read GetIntegerProp write SetIntegerProp stored False;
    property NotificationMode: Smallint index 510 read GetSmallintProp write SetSmallintProp stored False;
    property Logging: TOleBool index 514 read GetOleBoolProp write SetOleBoolProp stored False;
    property URL: string index 1001 read GetStringProp write SetStringProp stored False;
    property OnError: TSMTPError read FOnError write FOnError;
    property OnTimeout: TSMTPTimeout read FOnTimeout write FOnTimeout;
    property OnCancel: TNotifyEvent read FOnCancel write FOnCancel;
    property OnStateChanged: TSMTPStateChanged read FOnStateChanged write FOnStateChanged;
    property OnProtocolStateChanged: TSMTPProtocolStateChanged read FOnProtocolStateChanged write FOnProtocolStateChanged;
    property OnBusy: TSMTPBusy read FOnBusy write FOnBusy;
    property OnLog: TNotifyEvent read FOnLog write FOnLog;
    property OnDocInput: TSMTPDocInput read FOnDocInput write FOnDocInput;
    property OnReset: TNotifyEvent read FOnReset write FOnReset;
    property OnVerify: TNotifyEvent read FOnVerify write FOnVerify;
    property OnExpand: TNotifyEvent read FOnExpand write FOnExpand;
    property OnHelp: TNotifyEvent read FOnHelp write FOnHelp;
    property OnNoop: TNotifyEvent read FOnNoop write FOnNoop;
  end;

{ TTCP }

  TTCPError = procedure(Sender: TObject; Number: Smallint; var Description: string; Scode: Integer; const Source, HelpFile: string; HelpContext: Integer; var CancelDisplay: TOleBool) of object;
  TTCPDataArrival = procedure(Sender: TObject; bytesTotal: Integer) of object;
  TTCPConnectionRequest = procedure(Sender: TObject; requestID: Integer) of object;
  TTCPSendProgress = procedure(Sender: TObject; bytesSent, bytesRemaining: Integer) of object;

  TTCP = class(TOleControl)
  private
    FOnError: TTCPError;
    FOnDataArrival: TTCPDataArrival;
    FOnConnect: TNotifyEvent;
    FOnConnectionRequest: TTCPConnectionRequest;
    FOnClose: TNotifyEvent;
    FOnSendProgress: TTCPSendProgress;
    FOnSendComplete: TNotifyEvent;
  protected
    procedure InitControlData; override;
  public
    procedure AboutBox; stdcall;
    procedure Connect(const RemoteHost, RemotePort: Variant); stdcall;
    procedure Listen; stdcall;
    procedure Accept(requestID: Integer); stdcall;
    procedure SendData(const data: Variant); stdcall;
    procedure GetData(var data: Variant; const type_, maxLen: Variant); stdcall;
    procedure PeekData(var data: Variant; const type_, maxLen: Variant); stdcall;
    procedure Close; stdcall;
    property RemoteHostIP: string index 1001 read GetStringProp;
    property LocalHostName: string index 1002 read GetStringProp;
    property LocalIP: string index 1003 read GetStringProp;
    property SocketHandle: Integer index 1004 read GetIntegerProp;
    property State: Smallint index 503 read GetSmallintProp;
    property BytesReceived: Integer index 1101 read GetIntegerProp;
  published
    property RemoteHost: string index 0 read GetStringProp write SetStringProp stored False;
    property RemotePort: Integer index 502 read GetIntegerProp write SetIntegerProp stored False;
    property LocalPort: Integer index 1010 read GetIntegerProp write SetIntegerProp stored False;
    property OnError: TTCPError read FOnError write FOnError;
    property OnDataArrival: TTCPDataArrival read FOnDataArrival write FOnDataArrival;
    property OnConnect: TNotifyEvent read FOnConnect write FOnConnect;
    property OnConnectionRequest: TTCPConnectionRequest read FOnConnectionRequest write FOnConnectionRequest;
    property OnClose: TNotifyEvent read FOnClose write FOnClose;
    property OnSendProgress: TTCPSendProgress read FOnSendProgress write FOnSendProgress;
    property OnSendComplete: TNotifyEvent read FOnSendComplete write FOnSendComplete;
  end;

{ TUDP }
  
  TUDPError = procedure(Sender: TObject; Number: Smallint; var Description: string; Scode: Integer; const Source, HelpFile: string; HelpContext: Integer; var CancelDisplay: TOleBool) of object;
  TUDPDataArrival = procedure(Sender: TObject; bytesTotal: Integer) of object;

  TUDP = class(TOleControl)
  private
    FOnError: TUDPError;
    FOnDataArrival: TUDPDataArrival;
  protected
    procedure InitControlData; override;
  public
    procedure AboutBox; stdcall;
    procedure SendData(const data: Variant); stdcall;
    procedure GetData(var data: Variant; const type_: Variant); stdcall;
    property RemoteHostIP: string index 1001 read GetStringProp;
    property LocalHostName: string index 1002 read GetStringProp;
    property LocalIP: string index 1003 read GetStringProp;
    property SocketHandle: Integer index 1004 read GetIntegerProp;
  published
    property RemoteHost: string index 0 read GetStringProp write SetStringProp stored False;
    property RemotePort: Integer index 502 read GetIntegerProp write SetIntegerProp stored False;
    property LocalPort: Integer index 1010 read GetIntegerProp write SetIntegerProp stored False;
    property OnError: TUDPError read FOnError write FOnError;
    property OnDataArrival: TUDPDataArrival read FOnDataArrival write FOnDataArrival;
  end;

procedure Register;  
  
implementation

{$J+}

{ TFTP }

procedure TFTP.InitControlData;
const
  CEventDispIDs: array[0..26] of Integer = (
    $FFFFFDA0, $00000227, $00000228, $00000229, $0000022A, $0000022B,
    $0000022C, $000003F8, $000003F9, $0000002B, $0000002C, $0000002D,
    $0000002F, $00000030, $00000031, $00000036, $00000037, $00000038,
    $00000039, $0000003A, $0000003C, $0000003E, $0000003F, $00000040,
    $00000042, $00000044, $00000043);
  CControlData: TControlData = (
    ClassID: (
      D1:$B7FC354C;D2:$8CE7;D3:$11CF;D4:($97,$54,$00,$AA,$00,$C0,$09,$08));
    EventIID: (
      D1:$B7FC354B;D2:$8CE7;D3:$11CF;D4:($97,$54,$00,$AA,$00,$C0,$09,$08));
    EventCount: 27;
    EventDispIDs: @CEventDispIDs;
    LicenseKey: nil;
    Flags: $00000000);
begin
  ControlData := @CControlData;
end;

procedure TFTP.AboutBox;
const
  DispInfo: array[0..7] of Byte = ($D8,$FD,$FF,$FF,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.Cancel;
const
  DispInfo: array[0..7] of Byte = ($08,$02,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.Connect;
const
  DispInfo: array[0..9] of Byte = ($09,$02,$00,$00,$00,$01,$02,$00,$0C,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.Authenticate;
const
  DispInfo: array[0..9] of Byte = ($62,$02,$00,$00,$00,$01,$02,$00,$0C,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.SendDoc;
const
  DispInfo: array[0..12] of Byte = ($ED,$03,$00,$00,$00,$01,$05,$00,$0C,$0C,$0C,$0C,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.GetDoc;
const
  DispInfo: array[0..10] of Byte = ($EC,$03,$00,$00,$00,$01,$03,$00,$0C,$0C,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.Abort;
const
  DispInfo: array[0..7] of Byte = ($08,$00,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.Account;
const
  DispInfo: array[0..8] of Byte = ($09,$00,$00,$00,$00,$01,$01,$00,$08);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.ChangeDir;
const
  DispInfo: array[0..8] of Byte = ($0A,$00,$00,$00,$00,$01,$01,$00,$08);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.CreateDir;
const
  DispInfo: array[0..8] of Byte = ($0B,$00,$00,$00,$00,$01,$01,$00,$08);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.DeleteDir;
const
  DispInfo: array[0..8] of Byte = ($0C,$00,$00,$00,$00,$01,$01,$00,$08);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.DeleteFile;
const
  DispInfo: array[0..8] of Byte = ($0D,$00,$00,$00,$00,$01,$01,$00,$08);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.Quit;
const
  DispInfo: array[0..7] of Byte = ($0E,$00,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.Help;
const
  DispInfo: array[0..8] of Byte = ($10,$00,$00,$00,$00,$01,$01,$00,$08);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.Noop;
const
  DispInfo: array[0..7] of Byte = ($12,$00,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.mode;
const
  DispInfo: array[0..8] of Byte = ($13,$00,$00,$00,$00,$01,$01,$00,$02);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.Type_;
const
  DispInfo: array[0..8] of Byte = ($14,$00,$00,$00,$00,$01,$01,$00,$02);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.List;
const
  DispInfo: array[0..8] of Byte = ($15,$00,$00,$00,$00,$01,$01,$00,$08);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.NameList;
const
  DispInfo: array[0..8] of Byte = ($16,$00,$00,$00,$00,$01,$01,$00,$08);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.ParentDir;
const
  DispInfo: array[0..7] of Byte = ($17,$00,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.PrintDir;
const
  DispInfo: array[0..7] of Byte = ($18,$00,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.Execute;
const
  DispInfo: array[0..8] of Byte = ($19,$00,$00,$00,$00,$01,$01,$00,$08);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.Status;
const
  DispInfo: array[0..8] of Byte = ($1A,$00,$00,$00,$00,$01,$01,$00,$08);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.PutFile;
const
  DispInfo: array[0..9] of Byte = ($1B,$00,$00,$00,$00,$01,$02,$00,$08,$08);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.Reinitialize;
const
  DispInfo: array[0..7] of Byte = ($1C,$00,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.System;
const
  DispInfo: array[0..7] of Byte = ($1D,$00,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.GetFile;
const
  DispInfo: array[0..9] of Byte = ($1E,$00,$00,$00,$00,$01,$02,$00,$08,$08);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.PutFileUnique;
const
  DispInfo: array[0..8] of Byte = ($1F,$00,$00,$00,$00,$01,$01,$00,$08);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.Site;
const
  DispInfo: array[0..8] of Byte = ($20,$00,$00,$00,$00,$01,$01,$00,$08);
begin
  InvokeMethod(DispInfo, nil);
end;

function TFTP.Get_Timeout;
const
  DispInfo: array[0..8] of Byte = ($FB,$01,$00,$00,$03,$02,$01,$00,$02);
begin
  InvokeMethod(DispInfo, @Result);
end;

procedure TFTP.Set_Timeout;
const
  DispInfo: array[0..9] of Byte = ($FB,$01,$00,$00,$00,$04,$02,$00,$02,$03);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TFTP.Set_EnableTimer;
const
  DispInfo: array[0..9] of Byte = ($01,$02,$00,$00,$00,$04,$02,$00,$02,$0B);
begin
  InvokeMethod(DispInfo, nil);
end;

{ THTML }

procedure THTML.InitControlData;
const
  CEventDispIDs: array[0..12] of Integer = (
    $FFFFFDA0, $000003F8, $000003F9, $00000001, $00000002, $00000227,
    $00000004, $00000005, $00000006, $00000007, $00000008, $00000009,
    $0000000A);
  CControlData: TControlData = (
    ClassID: (
      D1:$B7FC355E;D2:$8CE7;D3:$11CF;D4:($97,$54,$00,$AA,$00,$C0,$09,$08));
    EventIID: (
      D1:$B7FC355D;D2:$8CE7;D3:$11CF;D4:($97,$54,$00,$AA,$00,$C0,$09,$08));
    EventCount: 13;
    EventDispIDs: @CEventDispIDs;
    LicenseKey: nil;
    Flags: $00000007);
begin
  ControlData := @CControlData;
end;

procedure THTML.AboutBox;
const
  DispInfo: array[0..7] of Byte = ($D8,$FD,$FF,$FF,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure THTML.RequestDoc;
const
  DispInfo: array[0..8] of Byte = ($22,$00,$00,$00,$00,$01,$01,$00,$08);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure THTML.RequestAllEmbedded;
const
  DispInfo: array[0..7] of Byte = ($23,$00,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure THTML.Cancel;
const
  DispInfo: array[0..8] of Byte = ($08,$02,$00,$00,$00,$01,$01,$00,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

{ THTTP }

procedure THTTP.InitControlData;
const
  CEventDispIDs: array[0..8] of Integer = (
    $FFFFFDA0, $00000227, $00000228, $00000229, $0000022A, $0000022B,
    $0000022C, $000003F8, $000003F9);
  CControlData: TControlData = (
    ClassID: (
      D1:$B7FC3563;D2:$8CE7;D3:$11CF;D4:($97,$54,$00,$AA,$00,$C0,$09,$08));
    EventIID: (
      D1:$B7FC3562;D2:$8CE7;D3:$11CF;D4:($97,$54,$00,$AA,$00,$C0,$09,$08));
    EventCount: 9;
    EventDispIDs: @CEventDispIDs;
    LicenseKey: nil;
    Flags: $00000000);
begin
  ControlData := @CControlData;
end;

procedure THTTP.AboutBox;
const
  DispInfo: array[0..7] of Byte = ($D8,$FD,$FF,$FF,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure THTTP.Cancel;
const
  DispInfo: array[0..7] of Byte = ($08,$02,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure THTTP.SendDoc;
const
  DispInfo: array[0..12] of Byte = ($ED,$03,$00,$00,$00,$01,$05,$00,$0C,$0C,$0C,$0C,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure THTTP.GetDoc;
const
  DispInfo: array[0..10] of Byte = ($EC,$03,$00,$00,$00,$01,$03,$00,$0C,$0C,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

function THTTP.Get_Timeout;
const
  DispInfo: array[0..8] of Byte = ($FB,$01,$00,$00,$03,$02,$01,$00,$02);
begin
  InvokeMethod(DispInfo, @Result);
end;

procedure THTTP.Set_Timeout;
const
  DispInfo: array[0..9] of Byte = ($FB,$01,$00,$00,$00,$04,$02,$00,$02,$03);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure THTTP.Set_EnableTimer;
const
  DispInfo: array[0..9] of Byte = ($01,$02,$00,$00,$00,$04,$02,$00,$02,$0B);
begin
  InvokeMethod(DispInfo, nil);
end;

{ TNNTP }

procedure TNNTP.InitControlData;
const
  CEventDispIDs: array[0..15] of Integer = (
    $FFFFFDA0, $00000227, $00000228, $00000229, $0000022A, $0000022B,
    $0000022C, $000003F8, $000003F9, $00000000, $00000001, $00000002,
    $00000003, $00000004, $00000005, $00000006);
  CControlData: TControlData = (
    ClassID: (
      D1:$B7FC3568;D2:$8CE7;D3:$11CF;D4:($97,$54,$00,$AA,$00,$C0,$09,$08));
    EventIID: (
      D1:$B7FC3567;D2:$8CE7;D3:$11CF;D4:($97,$54,$00,$AA,$00,$C0,$09,$08));
    EventCount: 16;
    EventDispIDs: @CEventDispIDs;
    LicenseKey: nil;
    Flags: $00000000);
begin
  ControlData := @CControlData;
end;

procedure TNNTP.AboutBox;
const
  DispInfo: array[0..7] of Byte = ($D8,$FD,$FF,$FF,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.Cancel;
const
  DispInfo: array[0..7] of Byte = ($08,$02,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.Connect;
const
  DispInfo: array[0..9] of Byte = ($09,$02,$00,$00,$00,$01,$02,$00,$0C,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.SendDoc;
const
  DispInfo: array[0..12] of Byte = ($ED,$03,$00,$00,$00,$01,$05,$00,$0C,$0C,$0C,$0C,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.GetDoc;
const
  DispInfo: array[0..10] of Byte = ($EC,$03,$00,$00,$00,$01,$03,$00,$0C,$0C,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.GetAdministrationFile;
const
  DispInfo: array[0..8] of Byte = ($05,$00,$00,$00,$00,$01,$01,$00,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.SelectGroup;
const
  DispInfo: array[0..8] of Byte = ($06,$00,$00,$00,$00,$01,$01,$00,$08);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.SetNextArticle;
const
  DispInfo: array[0..7] of Byte = ($07,$00,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.SetLastArticle;
const
  DispInfo: array[0..7] of Byte = ($08,$00,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.GetArticleNumbers;
const
  DispInfo: array[0..8] of Byte = ($09,$00,$00,$00,$00,$01,$01,$00,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.GetArticleHeaders;
const
  DispInfo: array[0..10] of Byte = ($0A,$00,$00,$00,$00,$01,$03,$00,$08,$0C,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.GetArticleByArticleNumber;
const
  DispInfo: array[0..8] of Byte = ($0B,$00,$00,$00,$00,$01,$01,$00,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.GetArticleByMessageID;
const
  DispInfo: array[0..8] of Byte = ($0C,$00,$00,$00,$00,$01,$01,$00,$08);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.GetHeaderByArticleNumber;
const
  DispInfo: array[0..8] of Byte = ($0D,$00,$00,$00,$00,$01,$01,$00,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.GetHeaderByMessageID;
const
  DispInfo: array[0..8] of Byte = ($0E,$00,$00,$00,$00,$01,$01,$00,$08);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.GetBodyByArticleNumber;
const
  DispInfo: array[0..8] of Byte = ($0F,$00,$00,$00,$00,$01,$01,$00,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.GetBodyByMessageID;
const
  DispInfo: array[0..8] of Byte = ($10,$00,$00,$00,$00,$01,$01,$00,$08);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.GetStatByArticleNumber;
const
  DispInfo: array[0..8] of Byte = ($11,$00,$00,$00,$00,$01,$01,$00,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.GetOverviewFormat;
const
  DispInfo: array[0..7] of Byte = ($12,$00,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.GetOverview;
const
  DispInfo: array[0..9] of Byte = ($13,$00,$00,$00,$00,$01,$02,$00,$0C,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.ListGroups;
const
  DispInfo: array[0..7] of Byte = ($14,$00,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.ListGroupDescriptions;
const
  DispInfo: array[0..7] of Byte = ($15,$00,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.ListNewGroups;
const
  DispInfo: array[0..8] of Byte = ($16,$00,$00,$00,$00,$01,$01,$00,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.Quit;
const
  DispInfo: array[0..7] of Byte = ($18,$00,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

function TNNTP.Get_Timeout;
const
  DispInfo: array[0..8] of Byte = ($FB,$01,$00,$00,$03,$02,$01,$00,$02);
begin
  InvokeMethod(DispInfo, @Result);
end;

procedure TNNTP.Set_Timeout;
const
  DispInfo: array[0..9] of Byte = ($FB,$01,$00,$00,$00,$04,$02,$00,$02,$03);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TNNTP.Set_EnableTimer;
const
  DispInfo: array[0..9] of Byte = ($01,$02,$00,$00,$00,$04,$02,$00,$02,$0B);
begin
  InvokeMethod(DispInfo, nil);
end;

{ TPOP }

procedure TPOP.InitControlData;
const
  CEventDispIDs: array[0..13] of Integer = (
    $FFFFFDA0, $00000227, $00000228, $00000229, $0000022A, $0000022B,
    $0000022C, $000003F9, $000009A1, $000009A2, $000009A3, $000009A4,
    $000009A6, $000009A7);
  CControlData: TControlData = (
    ClassID: (
      D1:$B7FC356E;D2:$8CE7;D3:$11CF;D4:($97,$54,$00,$AA,$00,$C0,$09,$08));
    EventIID: (
      D1:$B7FC356D;D2:$8CE7;D3:$11CF;D4:($97,$54,$00,$AA,$00,$C0,$09,$08));
    EventCount: 14;
    EventDispIDs: @CEventDispIDs;
    LicenseKey: nil;
    Flags: $00000000);
begin
  ControlData := @CControlData;
end;

procedure TPOP.AboutBox;
const
  DispInfo: array[0..7] of Byte = ($D8,$FD,$FF,$FF,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TPOP.Cancel;
const
  DispInfo: array[0..7] of Byte = ($08,$02,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TPOP.Connect;
const
  DispInfo: array[0..9] of Byte = ($09,$02,$00,$00,$00,$01,$02,$00,$0C,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TPOP.Authenticate;
const
  DispInfo: array[0..9] of Byte = ($62,$02,$00,$00,$00,$01,$02,$00,$0C,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TPOP.GetDoc;
const
  DispInfo: array[0..10] of Byte = ($EC,$03,$00,$00,$00,$01,$03,$00,$0C,$0C,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TPOP.MessageSize;
const
  DispInfo: array[0..8] of Byte = ($98,$09,$00,$00,$00,$01,$01,$00,$02);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TPOP.RetrieveMessage;
const
  DispInfo: array[0..8] of Byte = ($99,$09,$00,$00,$00,$01,$01,$00,$02);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TPOP.Delete;
const
  DispInfo: array[0..8] of Byte = ($9A,$09,$00,$00,$00,$01,$01,$00,$02);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TPOP.Reset;
const
  DispInfo: array[0..7] of Byte = ($9B,$09,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TPOP.Last;
const
  DispInfo: array[0..7] of Byte = ($9C,$09,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TPOP.Noop;
const
  DispInfo: array[0..7] of Byte = ($97,$09,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TPOP.TopMessage;
const
  DispInfo: array[0..8] of Byte = ($9D,$09,$00,$00,$00,$01,$01,$00,$02);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TPOP.Quit;
const
  DispInfo: array[0..7] of Byte = ($9E,$09,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TPOP.RefreshMessageCount;
const
  DispInfo: array[0..7] of Byte = ($C4,$09,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

function TPOP.Get_Timeout;
const
  DispInfo: array[0..8] of Byte = ($FB,$01,$00,$00,$03,$02,$01,$00,$02);
begin
  InvokeMethod(DispInfo, @Result);
end;

procedure TPOP.Set_Timeout;
const
  DispInfo: array[0..9] of Byte = ($FB,$01,$00,$00,$00,$04,$02,$00,$02,$03);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TPOP.Set_EnableTimer;
const
  DispInfo: array[0..9] of Byte = ($01,$02,$00,$00,$00,$04,$02,$00,$02,$0B);
begin
  InvokeMethod(DispInfo, nil);
end;

{ TSMTP }

procedure TSMTP.InitControlData;
const
  CEventDispIDs: array[0..12] of Integer = (
    $FFFFFDA0, $00000227, $00000228, $00000229, $0000022A, $0000022B,
    $0000022C, $000003F8, $0000012C, $0000012D, $0000012E, $0000012F,
    $00000130);
  CControlData: TControlData = (
    ClassID: (
      D1:$B7FC3573;D2:$8CE7;D3:$11CF;D4:($97,$54,$00,$AA,$00,$C0,$09,$08));
    EventIID: (
      D1:$B7FC3572;D2:$8CE7;D3:$11CF;D4:($97,$54,$00,$AA,$00,$C0,$09,$08));
    EventCount: 13;
    EventDispIDs: @CEventDispIDs;
    LicenseKey: nil;
    Flags: $00000000);
begin
  ControlData := @CControlData;
end;

procedure TSMTP.AboutBox;
const
  DispInfo: array[0..7] of Byte = ($D8,$FD,$FF,$FF,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TSMTP.Cancel;
const
  DispInfo: array[0..7] of Byte = ($08,$02,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TSMTP.Connect;
const
  DispInfo: array[0..9] of Byte = ($09,$02,$00,$00,$00,$01,$02,$00,$0C,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TSMTP.SendDoc;
const
  DispInfo: array[0..12] of Byte = ($ED,$03,$00,$00,$00,$01,$05,$00,$0C,$0C,$0C,$0C,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TSMTP.Reset;
const
  DispInfo: array[0..7] of Byte = ($C8,$00,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TSMTP.Verify;
const
  DispInfo: array[0..8] of Byte = ($C9,$00,$00,$00,$00,$01,$01,$00,$08);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TSMTP.Expand;
const
  DispInfo: array[0..8] of Byte = ($CA,$00,$00,$00,$00,$01,$01,$00,$08);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TSMTP.Help;
const
  DispInfo: array[0..8] of Byte = ($CB,$00,$00,$00,$00,$01,$01,$00,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TSMTP.Noop;
const
  DispInfo: array[0..7] of Byte = ($CC,$00,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TSMTP.Quit;
const
  DispInfo: array[0..7] of Byte = ($CD,$00,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

function TSMTP.Get_Timeout;
const
  DispInfo: array[0..8] of Byte = ($FB,$01,$00,$00,$03,$02,$01,$00,$02);
begin
  InvokeMethod(DispInfo, @Result);
end;

procedure TSMTP.Set_Timeout;
const
  DispInfo: array[0..9] of Byte = ($FB,$01,$00,$00,$00,$04,$02,$00,$02,$03);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TSMTP.Set_EnableTimer;
const
  DispInfo: array[0..9] of Byte = ($01,$02,$00,$00,$00,$04,$02,$00,$02,$0B);
begin
  InvokeMethod(DispInfo, nil);
end;

{ TTCP }

procedure TTCP.InitControlData;
const
  CEventDispIDs: array[0..6] of Integer = (
    $FFFFFDA0, $0000041B, $0000047F, $00000480, $00000483, $00000481,
    $00000482);
  CControlData: TControlData = (
    ClassID: (
      D1:$B7FC3584;D2:$8CE7;D3:$11CF;D4:($97,$54,$00,$AA,$00,$C0,$09,$08));
    EventIID: (
      D1:$B7FC3583;D2:$8CE7;D3:$11CF;D4:($97,$54,$00,$AA,$00,$C0,$09,$08));
    EventCount: 7;
    EventDispIDs: @CEventDispIDs;
    LicenseKey: nil;
    Flags: $00000000);
begin
  ControlData := @CControlData;
end;

procedure TTCP.AboutBox;
const
  DispInfo: array[0..7] of Byte = ($D8,$FD,$FF,$FF,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TTCP.Connect;
const
  DispInfo: array[0..9] of Byte = ($09,$02,$00,$00,$00,$01,$02,$00,$0C,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TTCP.Listen;
const
  DispInfo: array[0..7] of Byte = ($4E,$04,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TTCP.Accept;
const
  DispInfo: array[0..8] of Byte = ($4F,$04,$00,$00,$00,$01,$01,$00,$03);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TTCP.SendData;
const
  DispInfo: array[0..8] of Byte = ($F4,$03,$00,$00,$00,$01,$01,$00,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TTCP.GetData;
const
  DispInfo: array[0..10] of Byte = ($F3,$03,$00,$00,$00,$01,$03,$00,$8C,$0C,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TTCP.PeekData;
const
  DispInfo: array[0..10] of Byte = ($50,$04,$00,$00,$00,$01,$03,$00,$8C,$0C,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TTCP.Close;
const
  DispInfo: array[0..7] of Byte = ($51,$04,$00,$00,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

{ TUDP }

procedure TUDP.InitControlData;
const
  CEventDispIDs: array[0..1] of Integer = (
    $FFFFFDA0, $0000041B);
  CControlData: TControlData = (
    ClassID: (
      D1:$B7FC3587;D2:$8CE7;D3:$11CF;D4:($97,$54,$00,$AA,$00,$C0,$09,$08));
    EventIID: (
      D1:$B7FC3586;D2:$8CE7;D3:$11CF;D4:($97,$54,$00,$AA,$00,$C0,$09,$08));
    EventCount: 2;
    EventDispIDs: @CEventDispIDs;
    LicenseKey: nil;
    Flags: $00000000);
begin
  ControlData := @CControlData;
end;

procedure TUDP.AboutBox;
const
  DispInfo: array[0..7] of Byte = ($D8,$FD,$FF,$FF,$00,$01,$00,$00);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TUDP.SendData;
const
  DispInfo: array[0..8] of Byte = ($F4,$03,$00,$00,$00,$01,$01,$00,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure TUDP.GetData;
const
  DispInfo: array[0..9] of Byte = ($F3,$03,$00,$00,$00,$01,$02,$00,$8C,$0C);
begin
  InvokeMethod(DispInfo, nil);
end;

procedure Register;
begin
  RegisterComponents('Internet', [TFTP, THTML, THTTP, TNNTP, TPOP, TSMTP, TTCP, TUDP]);
end;


end.
 
